# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import List

from nemo_curator.services.conversation_formatter import ConversationFormatter


class NoFormat(ConversationFormatter):

    def format_conversation(self, conv: List[dict]) -> str:
        if len(conv) != 1:
            raise ValueError(
                "There must be exactly one turn in the conversation to use NoFormat"
            )

        turn = conv[0]

        if turn["role"] != "user":
            raise ValueError(
                "Conversation turn 0 is not 'user'. All even number turns should be."
            )

        return turn["content"]
